"""Tests of schema sub-types."""

import os

import fiona
from fiona.model import Feature


def test_read_bool_subtype(tmp_path):
    test_data = """{"type": "FeatureCollection", "features": [{"type": "Feature", "properties": {"bool": true, "not_bool": 1, "float": 42.5}, "geometry": null}]}"""
    path = tmp_path.joinpath("test_read_bool_subtype.geojson")

    with open(os.fspath(path), "w") as f:
        f.write(test_data)

    with fiona.open(path, "r") as src:
        feature = next(iter(src))

    assert type(feature["properties"]["bool"]) is bool
    assert isinstance(feature["properties"]["not_bool"], int)
    assert type(feature["properties"]["float"]) is float


def test_write_bool_subtype(tmp_path):
    path = tmp_path.joinpath("test_write_bool_subtype.geojson")

    schema = {
        "geometry": "Point",
        "properties": {
            "bool": "bool",
            "not_bool": "int",
            "float": "float",
        },
    }

    feature = Feature.from_dict(
        **{
            "geometry": None,
            "properties": {
                "bool": True,
                "not_bool": 1,
                "float": 42.5,
            },
        }
    )

    with fiona.open(path, "w", driver="GeoJSON", schema=schema) as dst:
        dst.write(feature)

    with open(os.fspath(path)) as f:
        data = f.read()

    assert """"bool": true""" in data
    assert """"not_bool": 1""" in data


def test_write_int16_subtype(tmp_path):
    path = tmp_path.joinpath("test_write_bool_subtype.gpkg")

    schema = {
        "geometry": "Point",
        "properties": {
            "a": "int",
            "b": "int16",
        },
    }

    feature = Feature.from_dict(
        **{
            "geometry": None,
            "properties": {
                "a": 1,
                "b": 2,
            },
        }
    )

    with fiona.open(path, "w", driver="GPKG", schema=schema) as colxn:
        colxn.write(feature)

    with fiona.open(path) as colxn:
        assert colxn.schema["properties"]["a"] == "int"
        assert colxn.schema["properties"]["b"] == "int16"



"""Fiona schema module."""

include "gdal.pxi"

import itertools
from typing import List

from fiona.errors import SchemaError


cdef class FionaIntegerType:
    names = ["int32"]
    type = int


cdef class FionaInt16Type:
    names = ["int16"]
    type = int


cdef class FionaBooleanType:
    names = ["bool"]
    type = bool


cdef class FionaInteger64Type:
    names = ["int", "int64"]
    type = int


cdef class FionaRealType:
    names = ["float", "float64"]
    type = float


cdef class FionaStringType:
    names = ["str"]
    type = str


cdef class FionaBinaryType:
    names = ["bytes"]
    type = bytes


cdef class FionaStringListType:
    names = ["List[str]", "list[str]"]
    type = List[str]


cdef class FionaJSONType:
    names = ["json"]
    type = str


cdef class FionaDateType:
    """Dates without time."""
    names = ["date"]
    type = str


cdef class FionaTimeType:
    """Times without dates."""
    names = ["time"]
    type = str


cdef class FionaDateTimeType:
    """Dates and times."""
    names = ["datetime"]
    type = str


FIELD_TYPES_MAP2_REV = {
    (OFTInteger, OFSTNone): FionaIntegerType,
    (OFTInteger, OFSTBoolean): FionaBooleanType,
    (OFTInteger, OFSTInt16): FionaInt16Type,
    (OFTInteger64, OFSTNone): FionaInteger64Type,
    (OFTReal, OFSTNone): FionaRealType,
    (OFTString, OFSTNone): FionaStringType,
    (OFTDate, OFSTNone): FionaDateType,
    (OFTTime, OFSTNone): FionaTimeType,
    (OFTDateTime, OFSTNone): FionaDateTimeType,
    (OFTBinary, OFSTNone): FionaBinaryType,
    (OFTStringList, OFSTNone): FionaStringListType,
    (OFTString, OFSTJSON): FionaJSONType,
}
FIELD_TYPES_MAP2 = {v: k for k, v in FIELD_TYPES_MAP2_REV.items()}
FIELD_TYPES_NAMES = list(itertools.chain.from_iterable((k.names for k in FIELD_TYPES_MAP2)))
NAMED_FIELD_TYPES = {n: k for k in FIELD_TYPES_MAP2 for n in k.names}


def normalize_field_type(ftype):
    """Normalize free form field types to an element of FIELD_TYPES

    Parameters
    ----------
    ftype : str
        A type:width format like 'int:9' or 'str:255'

    Returns
    -------
    str
        An element from FIELD_TYPES
    """
    if ftype in FIELD_TYPES_NAMES:
        return ftype
    elif ftype.startswith('int'):
        width = int((ftype.split(':')[1:] or ['0'])[0])
        if width == 0 or width >= 10:
            return 'int64'
        else:
            return 'int32'
    elif ftype.startswith('str'):
        return 'str'
    elif ftype.startswith('float'):
        return 'float'
    else:
        raise SchemaError(f"Unknown field type: {ftype}")


# Fiona field type names indexed by their major OGR integer field type.
# This data is deprecated, no longer used by the project and is left
# only for other projects that import it.
FIELD_TYPES = [
    'int32',        # OFTInteger, Simple 32bit integer
    None,           # OFTIntegerList, List of 32bit integers
    'float',        # OFTReal, Double Precision floating point
    None,           # OFTRealList, List of doubles
    'str',          # OFTString, String of UTF-8 chars
    'List[str]',    # OFTStringList, Array of strings
    None,           # OFTWideString, deprecated
    None,           # OFTWideStringList, deprecated
    'bytes',        # OFTBinary, Raw Binary data
    'date',         # OFTDate, Date
    'time',         # OFTTime, Time
    'datetime',     # OFTDateTime, Date and Time
    'int64',        # OFTInteger64, Single 64bit integer
    None            # OFTInteger64List, List of 64bit integers
]

# Mapping of Fiona field type names to Python types.
# This data is deprecated, no longer used by the project and is left
# only for other projects that import it.
FIELD_TYPES_MAP = {
    'int32': int,
    'float': float,
    'str': str,
    'date': FionaDateType,
    'time': FionaTimeType,
    'datetime': FionaDateTimeType,
    'bytes': bytes,
    'int64': int,
    'int': int,
    'List[str]': List[str],
}
FIELD_TYPES_MAP_REV = dict([(v, k) for k, v in FIELD_TYPES_MAP.items()])
FIELD_TYPES_MAP_REV[int] = 'int'





"""Tests of the Python opener VSI plugin."""

import io
import os

import fsspec
import pytest

import fiona
from fiona.model import Feature


def test_opener_io_open(path_grenada_geojson):
    """Use io.open as opener."""
    with fiona.open(path_grenada_geojson, opener=io.open) as colxn:
        profile = colxn.profile
        assert profile["driver"] == "GeoJSON"
        assert len(colxn) == 1


def test_opener_fsspec_zip_fs():
    """Use fsspec zip filesystem as opener."""
    fs = fsspec.filesystem("zip", fo="tests/data/coutwildrnp.zip")
    with fiona.open("coutwildrnp.shp", opener=fs) as colxn:
        profile = colxn.profile
        assert profile["driver"] == "ESRI Shapefile"
        assert len(colxn) == 67
        assert colxn.schema["geometry"] == "Polygon"
        assert "AGBUR" in colxn.schema["properties"]


def test_opener_fsspec_zip_http_fs():
    """Use fsspec zip+http filesystem as opener."""
    fs = fsspec.filesystem(
        "zip",
        target_protocol="http",
        fo="https://github.com/Toblerity/Fiona/files/11151652/coutwildrnp.zip",
    )
    with fiona.open("coutwildrnp.shp", opener=fs) as colxn:
        profile = colxn.profile
        assert profile["driver"] == "ESRI Shapefile"
        assert len(colxn) == 67
        assert colxn.schema["geometry"] == "Polygon"
        assert "AGBUR" in colxn.schema["properties"]


def test_opener_tiledb_file():
    """Use tiledb vfs as opener."""
    tiledb = pytest.importorskip("tiledb")
    fs = tiledb.VFS()
    with fiona.open("tests/data/coutwildrnp.shp", opener=fs) as colxn:
        profile = colxn.profile
        assert profile["driver"] == "ESRI Shapefile"
        assert len(colxn) == 67
        assert colxn.schema["geometry"] == "Polygon"
        assert "AGBUR" in colxn.schema["properties"]


def test_opener_fsspec_fs_write(tmp_path):
    """Write a feature via an fsspec fs opener."""
    schema = {"geometry": "Point", "properties": {"zero": "int"}}
    feature = Feature.from_dict(
        **{
            "geometry": {"type": "Point", "coordinates": (0, 0)},
            "properties": {"zero": "0"},
        }
    )
    fs = fsspec.filesystem("file")
    outputfile = tmp_path.joinpath("test.shp")

    with fiona.open(
        str(outputfile),
        "w",
        driver="ESRI Shapefile",
        schema=schema,
        crs="OGC:CRS84",
        opener=fs,
    ) as collection:
        collection.write(feature)
        assert len(collection) == 1
        assert collection.crs == "OGC:CRS84"


def test_threads_context():
    import io
    from threading import Thread


    def target():
        with fiona.open("tests/data/coutwildrnp.shp", opener=io.open) as colxn:
            print(colxn.profile)
            assert len(colxn) == 67


    thread = Thread(target=target)
    thread.start()
    thread.join()


def test_overwrite(data):
    """Opener can overwrite data."""
    schema = {"geometry": "Point", "properties": {"zero": "int"}}
    feature = Feature.from_dict(
        **{
            "geometry": {"type": "Point", "coordinates": (0, 0)},
            "properties": {"zero": "0"},
        }
    )
    fs = fsspec.filesystem("file")
    outputfile = os.path.join(str(data), "coutwildrnp.shp")

    with fiona.open(
        str(outputfile),
        "w",
        driver="ESRI Shapefile",
        schema=schema,
        crs="OGC:CRS84",
        opener=fs,
    ) as collection:
        collection.write(feature)
        assert len(collection) == 1
        assert collection.crs == "OGC:CRS84"


def test_opener_fsspec_zip_fs_listlayers():
    """Use fsspec zip filesystem as opener for listlayers()."""
    fs = fsspec.filesystem("zip", fo="tests/data/coutwildrnp.zip")
    assert fiona.listlayers("coutwildrnp.shp", opener=fs) == ["coutwildrnp"]


def test_opener_fsspec_zip_fs_listdir():
    """Use fsspec zip filesystem as opener for listdir()."""
    fs = fsspec.filesystem("zip", fo="tests/data/coutwildrnp.zip")
    listing = fiona.listdir("/", opener=fs)
    assert len(listing) == 4
    assert set(
        ["coutwildrnp.shp", "coutwildrnp.dbf", "coutwildrnp.shx", "coutwildrnp.prj"]
    ) & set(listing)


def test_opener_fsspec_file_fs_listdir():
    """Use fsspec file filesystem as opener for listdir()."""
    fs = fsspec.filesystem("file")
    listing = fiona.listdir("tests/data", opener=fs)
    assert len(listing) >= 33
    assert set(
        ["coutwildrnp.shp", "coutwildrnp.dbf", "coutwildrnp.shx", "coutwildrnp.prj"]
    ) & set(listing)


def test_opener_fsspec_file_remove(data):
    """Opener can remove data."""
    fs = fsspec.filesystem("file")
    listing = fiona.listdir(str(data), opener=fs)
    assert len(listing) == 4
    outputfile = os.path.join(str(data), "coutwildrnp.shp")
    fiona.remove(outputfile)
    listing = fiona.listdir(str(data), opener=fs)
    assert len(listing) == 0
    assert not set(
        ["coutwildrnp.shp", "coutwildrnp.dbf", "coutwildrnp.shx", "coutwildrnp.prj"]
    ) & set(listing)


def test_opener_tiledb_vfs_listdir():
    """Use tiledb VFS as opener for listdir()."""
    tiledb = pytest.importorskip("tiledb")
    fs = tiledb.VFS()
    listing = fiona.listdir("tests/data", opener=fs)
    assert len(listing) >= 33
    assert set(
        ["coutwildrnp.shp", "coutwildrnp.dbf", "coutwildrnp.shx", "coutwildrnp.prj"]
    ) & set(listing)


def test_opener_interface():
    """Demonstrate implementation of a custom opener."""
    import pathlib
    from fiona.abc import FileContainer

    class CustomContainer:
        """GDAL's VSI ReadDir() uses 5 of FileContainer's methods."""
        def isdir(self, path):
            return pathlib.Path(path).is_dir()

        def isfile(self, path):
            return pathlib.Path(path).is_file()

        def ls(self, path):
            return list(pathlib.Path(path).iterdir())

        def mtime(self, path):
            return pathlib.Path(path).stat().st_mtime

        def size(self, path):
            return pathlib.Path(path).stat().st_size

    FileContainer.register(CustomContainer)

    listing = fiona.listdir("tests/data", opener=CustomContainer())
    assert len(listing) >= 33
    assert set(
        ["coutwildrnp.shp", "coutwildrnp.dbf", "coutwildrnp.shx", "coutwildrnp.prj"]
    ) & set(listing)



"""_path tests."""

import sys

from fiona._path import _parse_path, _vsi_path


def test_parse_zip_windows(monkeypatch):
    """Parse a zip+ Windows path."""
    monkeypatch.setattr(sys, "platform", "win32")
    path = _parse_path("zip://D:\\a\\Fiona\\Fiona\\tests\\data\\coutwildrnp.zip!coutwildrnp.shp")
    vsi_path = _vsi_path(path)
    assert vsi_path.startswith("/vsizip/D")
    assert vsi_path.endswith("coutwildrnp.zip/coutwildrnp.shp")


def test_parse_zip_windows(monkeypatch):
    """Parse a tar+ Windows path."""
    monkeypatch.setattr(sys, "platform", "win32")
    path = _parse_path("tar://D:\\a\\Fiona\\Fiona\\tests\\data\\coutwildrnp.tar!testing/coutwildrnp.shp")
    vsi_path = _vsi_path(path)
    assert vsi_path.startswith("/vsitar/D")
    assert vsi_path.endswith("coutwildrnp.tar/testing/coutwildrnp.shp")



# CLI tests

from click.testing import CliRunner

from fiona.fio.main import main_group  # type: ignore
import pytest  # type: ignore


def test_map_count():
    """fio-map prints correct number of results."""
    with open("tests/data/trio.seq") as seq:
        data = seq.read()

    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ["map", "centroid (buffer g 1.0)"],
        input=data,
    )

    assert result.exit_code == 0
    assert result.output.count('"type": "Point"') == 3


@pytest.mark.parametrize("raw_opt", ["--raw", "-r"])
def test_reduce_area(raw_opt):
    """Reduce features to their (raw) area."""
    with open("tests/data/trio.seq") as seq:
        data = seq.read()

    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ["reduce", raw_opt, "area (unary_union c) :projected false"],
        input=data,
    )
    assert result.exit_code == 0
    assert 0 < float(result.output) < 1e-5


def test_reduce_union():
    """Reduce features to one single feature."""
    with open("tests/data/trio.seq") as seq:
        data = seq.read()

    # Define our reduce command using a mkdocs snippet.
    arg = """
    --8<-- [start:reduce]
    unary_union c
    --8<-- [end:reduce]
    """.splitlines()[
        2
    ].strip()

    runner = CliRunner()
    result = runner.invoke(main_group, ["reduce", arg], input=data)
    assert result.exit_code == 0
    assert result.output.count('"type": "Polygon"') == 1
    assert result.output.count('"type": "LineString"') == 1
    assert result.output.count('"type": "GeometryCollection"') == 1


def test_reduce_union_zip_properties():
    """Reduce features to one single feature, zipping properties."""
    with open("tests/data/trio.seq") as seq:
        data = seq.read()

    runner = CliRunner()
    result = runner.invoke(
        main_group, ["reduce", "--zip-properties", "unary_union c"], input=data
    )
    assert result.exit_code == 0
    assert result.output.count('"type": "Polygon"') == 1
    assert result.output.count('"type": "LineString"') == 1
    assert result.output.count('"type": "GeometryCollection"') == 1
    assert (
        """"name": ["Le ch\\u00e2teau d\'eau", "promenade du Peyrou"]"""
        in result.output
    )


def test_filter():
    """Filter features by distance."""
    with open("tests/data/trio.seq") as seq:
        data = seq.read()

    # Define our reduce command using a mkdocs snippet.
    arg = """
    --8<-- [start:filter]
    < (distance g (Point 4 43)) 62.5E3
    --8<-- [end:filter]
    """.splitlines()[
        2
    ].strip()

    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ["filter", arg],
        input=data,
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert result.output.count('"type": "Polygon"') == 1


@pytest.mark.parametrize("opts", [["--no-input", "--raw"], ["-rn"]])
def test_map_no_input(opts):
    runner = CliRunner()
    result = runner.invoke(main_group, ["map"] + opts + ["(Point 4 43)"])
    assert result.exit_code == 0
    assert result.output.count('"type": "Point"') == 1



# Fiona build script.

import logging
import os
import shutil
import subprocess
import sys

from setuptools import setup
from setuptools.extension import Extension


# Ensure minimum version of Python is running
if sys.version_info[0:2] < (3, 6):
    raise RuntimeError('Fiona requires Python>=3.6')

# Use Cython if available.
try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None


def check_output(cmd):
    return subprocess.check_output(cmd).decode('utf')


def copy_data_tree(datadir, destdir):
    try:
        shutil.rmtree(destdir)
    except OSError:
        pass
    shutil.copytree(datadir, destdir)


# Building Fiona requires options that can be obtained from GDAL's gdal-config
# program or can be specified using setup arguments. The latter override the
# former.
#
# A GDAL API version is strictly required. Without this the setup script
# cannot know whether to use the GDAL version 1 or 2 source files. The GDAL
# API version can be specified in 2 ways.
#
# 1. By the gdal-config program, optionally pointed to by GDAL_CONFIG
# 2. By a GDAL_VERSION environment variable. This overrides number 1.


include_dirs = []
library_dirs = []
libraries = []
extra_link_args = []
gdal_output = [None for i in range(4)]
gdalversion = None
language = None
gdal_major_version = 0
gdal_minor_version = 0

if 'clean' not in sys.argv:
    try:
        gdal_config = os.environ.get('GDAL_CONFIG', 'gdal-config')
        for i, flag in enumerate(
                ["--cflags", "--libs", "--datadir", "--version"]):
            gdal_output[i] = check_output([gdal_config, flag]).strip()
        for item in gdal_output[0].split():
            if item.startswith("-I"):
                include_dirs.extend(item[2:].split(":"))
        for item in gdal_output[1].split():
            if item.startswith("-L"):
                library_dirs.extend(item[2:].split(":"))
            elif item.startswith("-l"):
                libraries.append(item[2:])
            else:
                # e.g. -framework GDAL
                extra_link_args.append(item)
        gdalversion = gdal_output[3]
        if gdalversion:
            logging.info("GDAL API version obtained from gdal-config: %s",
                         gdalversion)

    except Exception as e:
        if os.name == "nt":
            logging.info("Building on Windows requires extra options to"
                         " setup.py to locate needed GDAL files.\nMore"
                         " information is available in the README.")
        else:
            logging.warning("Failed to get options via gdal-config: %s", str(e))

    # Get GDAL API version from environment variable.
    if 'GDAL_VERSION' in os.environ:
        gdalversion = os.environ['GDAL_VERSION']
        logging.info("GDAL API version obtained from environment: %s",
                     gdalversion)

    # Get GDAL API version from the command line if specified there.
    if '--gdalversion' in sys.argv:
        index = sys.argv.index('--gdalversion')
        sys.argv.pop(index)
        gdalversion = sys.argv.pop(index)
        logging.info("GDAL API version obtained from command line option: %s",
                     gdalversion)

    if not gdalversion:
        logging.fatal("A GDAL API version must be specified. Provide a path "
                      "to gdal-config using a GDAL_CONFIG environment "
                      "variable or use a GDAL_VERSION environment variable.")
        sys.exit(1)

    if os.environ.get('PACKAGE_DATA'):
        destdir = 'fiona/gdal_data'
        if gdal_output[2]:
            logging.info(f"Copying gdal data from {gdal_output[2]}")
            copy_data_tree(gdal_output[2], destdir)
        else:
            # check to see if GDAL_DATA is defined
            gdal_data = os.environ.get('GDAL_DATA', None)
            if gdal_data:
                logging.info(f"Copying gdal data from {gdal_data}")
                copy_data_tree(gdal_data, destdir)

        # Conditionally copy PROJ DATA.
        projdatadir = os.environ.get('PROJ_DATA', os.environ.get('PROJ_LIB', '/usr/local/share/proj'))
        if os.path.exists(projdatadir):
            logging.info(f"Copying proj data from {projdatadir}")
            copy_data_tree(projdatadir, 'fiona/proj_data')

    if "--cython-language" in sys.argv:
        index = sys.argv.index("--cython-language")
        sys.argv.pop(index)
        language = sys.argv.pop(index).lower()

    gdal_version_parts = gdalversion.split('.')
    gdal_major_version = int(gdal_version_parts[0])
    gdal_minor_version = int(gdal_version_parts[1])

    if (gdal_major_version, gdal_minor_version) < (3, 1):
        raise SystemExit(
            "ERROR: GDAL >= 3.1 is required for fiona. "
            "Please upgrade GDAL."
        )

    logging.info("GDAL version major=%r minor=%r", gdal_major_version,
                 gdal_minor_version)

compile_time_env = {
    "CTE_GDAL_MAJOR_VERSION": gdal_major_version,
    "CTE_GDAL_MINOR_VERSION": gdal_minor_version,
}

ext_options = dict(
    include_dirs=include_dirs,
    library_dirs=library_dirs,
    libraries=libraries,
    extra_link_args=extra_link_args,
    cython_compile_time_env=compile_time_env,
)

# Enable coverage for cython pyx files.
if os.environ.get('CYTHON_COVERAGE'):
    from Cython.Compiler.Options import get_directive_defaults
    directive_defaults = get_directive_defaults()
    directive_defaults['linetrace'] = True
    directive_defaults['binding'] = True

    ext_options.update(dict(
        define_macros=[("CYTHON_TRACE_NOGIL", "1")]))

# GDAL 2.3+ requires C++11

if language == "c++":
    ext_options["language"] = "c++"
    if sys.platform != "win32":
        ext_options["extra_compile_args"] = ["-std=c++11"]

ext_options_cpp = ext_options.copy()
if sys.platform != "win32":
    ext_options_cpp["extra_compile_args"] = ["-std=c++11"]


# Define the extension modules.
ext_modules = []

if "clean" not in sys.argv:
    # When building from a repo, Cython is required.
    logging.info("MANIFEST.in found, presume a repo, cythonizing...")
    if not cythonize:
        raise SystemExit(
            "Cython.Build.cythonize not found. "
            "Cython is required to build fiona."
        )

    ext_modules = cythonize(
        [
            Extension("fiona._geometry", ["fiona/_geometry.pyx"], **ext_options),
            Extension("fiona._vsiopener", ["fiona/_vsiopener.pyx"], **ext_options),
            Extension("fiona.schema", ["fiona/schema.pyx"], **ext_options),
            Extension("fiona._transform", ["fiona/_transform.pyx"], **ext_options_cpp),
            Extension("fiona.crs", ["fiona/crs.pyx"], **ext_options),
            Extension("fiona._env", ["fiona/_env.pyx"], **ext_options),
            Extension("fiona._err", ["fiona/_err.pyx"], **ext_options),
            Extension("fiona.ogrext", ["fiona/ogrext.pyx"], **ext_options),
        ],
        compiler_directives={"language_level": "3"},
        compile_time_env=compile_time_env,
    )

# Include these files for binary wheels
fiona_package_data = ['gdal.pxi', '*.pxd']

if os.environ.get('PACKAGE_DATA'):
    fiona_package_data.extend([
        'gdal_data/*',
        'proj_data/*',
        '.libs/*',
        '.libs/licenses/*',
    ])

# See pyproject.toml for project metadata
setup(
    name="Fiona",  # need by GitHub dependency graph
    package_data={"fiona": fiona_package_data},
    ext_modules=ext_modules,
)



# distutils: language = c++
#
# Coordinate and geometry transformations.

include "gdal.pxi"

import logging
import warnings
from collections import UserDict

from fiona cimport _cpl, _csl, _geometry
from fiona.crs cimport OGRSpatialReferenceH, osr_set_traditional_axis_mapping_strategy

from fiona.compat import DICT_TYPES
from fiona.crs import CRS
from fiona.errors import TransformError
from fiona.model import Geometry


cdef extern from "ogr_geometry.h" nogil:

    cdef cppclass OGRGeometry:
        pass

    cdef cppclass OGRGeometryFactory:
        void * transformWithOptions(void *geom, void *ct, char **options)


cdef extern from "ogr_spatialref.h":

    cdef cppclass OGRCoordinateTransformation:
        pass


log = logging.getLogger(__name__)
class NullHandler(logging.Handler):
    def emit(self, record):
        pass
log.addHandler(NullHandler())


cdef void *_crs_from_crs(object crs):
    cdef char *wkt_c = NULL
    cdef OGRSpatialReferenceH osr = NULL
    osr = OSRNewSpatialReference(NULL)

    if osr == NULL:
        raise ValueError("NULL spatial reference")

    params = []

    wkt = CRS.from_user_input(crs).to_wkt()

    wkt_b = wkt.encode('utf-8')
    wkt_c = wkt_b
    OSRSetFromUserInput(osr, wkt_c)

    osr_set_traditional_axis_mapping_strategy(osr)
    return osr


def _transform(src_crs, dst_crs, xs, ys):
    cdef double *x
    cdef double *y
    cdef char *proj_c = NULL
    cdef OGRSpatialReferenceH src = NULL
    cdef OGRSpatialReferenceH dst = NULL
    cdef void *transform = NULL
    cdef int i

    assert len(xs) == len(ys)

    src = _crs_from_crs(src_crs)
    dst = _crs_from_crs(dst_crs)

    n = len(xs)
    x = <double *>_cpl.CPLMalloc(n*sizeof(double))
    y = <double *>_cpl.CPLMalloc(n*sizeof(double))
    for i in range(n):
        x[i] = xs[i]
        y[i] = ys[i]

    transform = OCTNewCoordinateTransformation(src, dst)
    res = OCTTransform(transform, n, x, y, NULL)

    res_xs = [0]*n
    res_ys = [0]*n

    for i in range(n):
        res_xs[i] = x[i]
        res_ys[i] = y[i]

    _cpl.CPLFree(x)
    _cpl.CPLFree(y)
    OCTDestroyCoordinateTransformation(transform)
    OSRRelease(src)
    OSRRelease(dst)
    return res_xs, res_ys


cdef object _transform_single_geom(
    object single_geom,
    OGRGeometryFactory *factory,
    void *transform,
    char **options,
):
    """Transform a single geometry."""
    cdef void *src_ogr_geom = NULL
    cdef void *dst_ogr_geom = NULL

    src_ogr_geom = _geometry.OGRGeomBuilder().build(single_geom)
    dst_ogr_geom = factory.transformWithOptions(
                    <const OGRGeometry *>src_ogr_geom,
                    <OGRCoordinateTransformation *>transform,
                    options)

    if dst_ogr_geom == NULL and CPLGetConfigOption("OGR_ENABLE_PARTIAL_REPROJECTION", "OFF") != b"ON":
        raise TransformError(
            "Full reprojection failed. To enable partial reprojection set OGR_ENABLE_PARTIAL_REPROJECTION=True"
        )
    else:
        out_geom = _geometry.GeomBuilder().build(dst_ogr_geom)
        _geometry.OGR_G_DestroyGeometry(dst_ogr_geom)

    if src_ogr_geom != NULL:
        _geometry.OGR_G_DestroyGeometry(src_ogr_geom)

    return out_geom


def _transform_geom(src_crs, dst_crs, geom, antimeridian_cutting, antimeridian_offset, precision):
    """Return transformed geometries.

    """
    cdef char *proj_c = NULL
    cdef char *key_c = NULL
    cdef char *val_c = NULL
    cdef char **options = NULL
    cdef OGRSpatialReferenceH src = NULL
    cdef OGRSpatialReferenceH dst = NULL
    cdef void *transform = NULL
    cdef OGRGeometryFactory *factory = NULL

    if not all([src_crs, dst_crs]):
        raise RuntimeError("Must provide a source and destination CRS.")

    src = _crs_from_crs(src_crs)
    dst = _crs_from_crs(dst_crs)
    transform = OCTNewCoordinateTransformation(src, dst)

    # Transform options.
    options = _csl.CSLSetNameValue(
        options,
        "DATELINEOFFSET",
        str(antimeridian_offset).encode('utf-8')
    )

    if antimeridian_cutting:
        options = _csl.CSLSetNameValue(options, "WRAPDATELINE", "YES")

    factory = new OGRGeometryFactory()

    if isinstance(geom, Geometry):
        out_geom = recursive_round(
            _transform_single_geom(geom, factory, transform, options), precision)
    else:
        out_geom = [
            recursive_round(
                _transform_single_geom(single_geom, factory, transform, options),
                precision,
            )
            for single_geom in geom
        ]

    OCTDestroyCoordinateTransformation(transform)

    if options != NULL:
        _csl.CSLDestroy(options)

    OSRRelease(src)
    OSRRelease(dst)

    return out_geom


def recursive_round(obj, precision):
    """Recursively round coordinates."""
    if precision < 0:
        return obj
    if getattr(obj, 'geometries', None):
        return Geometry(geometries=[recursive_round(part, precision) for part in obj.geometries])
    elif getattr(obj, 'coordinates', None):
        return Geometry(coordinates=[recursive_round(part, precision) for part in obj.coordinates])
    if isinstance(obj, (int, float)):
        return round(obj, precision)
    else:
        return [recursive_round(part, precision) for part in obj]
